import json
import datetime

import pytest
import numpy as np
import pandas as pd
import xarray as xr

from wave_venture import serializer


@pytest.mark.parametrize(
    "input_,output",
    [
        [
            5,
            5,
        ],
        [
            "hello",
            "hello",
        ],
        [
            {"hello": "world"},
            {"hello": "world"},
        ],
        [
            datetime.datetime(2000, 1, 1),
            {"_type": "datetime", "epoch": 946684800.0},
        ],
        [
            [datetime.datetime(2000, 1, 1)],
            [{"_type": "datetime", "epoch": 946684800.0}],
        ],
        [
            np.datetime64("2000-01-01"),
            {"_type": "datetime", "epoch": 946684800.0},
        ],
        [
            np.array(np.datetime64("2000-01-01")),
            {"_type": "datetime", "epoch": 946684800.0},
        ],
        [
            np.array([np.datetime64("2000-01-01")]),
            [{"_type": "datetime", "epoch": 946684800.0}],
        ],
        [
            xr.DataArray(datetime.datetime(2000, 1, 1)),
            {"_type": "datetime", "epoch": 946684800.0},
        ],
        [
            xr.DataArray([datetime.datetime(2000, 1, 1)]),
            {
                "_type": "array",
                "fields": {
                    "unnamed": {
                        "epochs": [946684800.0],
                        "_type": "datetime"
                    }
                },
                "dimensions": [],
                "indices": {},
            },
        ],
        [
            np.array([np.datetime64("2000-01-01")]),
            [{"_type": "datetime", "epoch": 946684800.0}]
        ],
        [
            pd.Series([np.datetime64("2000-01-01")]),
            {
                "_type": "array",
                "fields": {
                    "unnamed": {
                        "epochs": [946684800.0],
                        "_type": "datetime"
                    }
                },
                "dimensions": ["index"],
                "indices": {
                    "index": {"values": [0], "_type": "number"}
                },
            },
        ],
    ],
)
def test_encoder(input_, output):
    candidate = json.loads(json.dumps(input_, cls=serializer.Encoder))
    assert candidate == output



import os
import json
import tempfile
import itertools
import subprocess

import numpy as np
import pandas as pd
import xarray as xr

from PIL import Image

from wave_venture import serializer
from wave_venture import config as cfg


HERE = os.path.dirname(os.path.realpath(__file__))
WATERMARK_PATH = os.path.abspath(
    os.path.join(HERE, "resources", "wave-venture-logo.png"),
)
DEFAULT = object()


def _apply_watermark(
    *,
    image_path,
    save_path=None,
    watermark_path,
    position="bottom-left",
    scale=1,
    offset=(0, 0),
):
    watermark = Image.open(watermark_path)
    wm_width, wm_height = watermark.size
    watermark = watermark.resize(
        (int(wm_width // (1 / scale)), int(wm_height // (1 / scale)))
    )
    wm_width, wm_height = watermark.size

    base_image = Image.open(image_path)
    image_width, image_height = base_image.size

    position_options = {
        "top-left": (0, 0),
        "top-right": (image_width - wm_width, 0),
        "bottom-left": (0, image_height - wm_height),
        "bottom-right": (image_width - wm_width, image_height - wm_height),
    }

    try:
        position_anchor = position_options[position]
    except KeyError:
        raise ValueError(f"position should be one of: {position_options}")

    offset_directions = {
        "top-left": (1, 1),
        "top-right": (-1, 1),
        "bottom-left": (1, -1),
        "bottom-right": (-1, -1),
    }

    position = (
        position_anchor[0] + (offset[0] * offset_directions[position][0]),
        position_anchor[1] + (offset[1] * offset_directions[position][1]),
    )

    transparent = Image.new("RGBA", (image_width, image_height), (0, 0, 0, 0))
    transparent.paste(base_image, (0, 0))
    transparent.paste(watermark, position, mask=watermark)
    transparent.save(save_path or image_path)


def _normalise_data(data):
    if not isinstance(data, list):
        raise ValueError("data must be a list.")

    out = []

    for entry in data:
        if isinstance(entry, list):
            entry = xr.DataArray(entry, coords={"index": range(len(entry))})
        elif isinstance(entry, np.ndarray):
            entry = xr.DataArray(entry, coords={"index": range(len(entry))})
        elif isinstance(entry, pd.Series):
            idx_name = entry.index.name or ""
            entry = xr.DataArray(entry.values, coords={idx_name: entry.index})
        elif isinstance(entry, pd.DataFrame):
            idx_name = entry.index.name or ""
            data = {c: ([idx_name], entry[c].values) for c in entry.columns}
            entry = xr.Dataset(data, coords={idx_name: entry.index})
        elif isinstance(entry, xr.DataArray):
            pass
        elif isinstance(entry, xr.Dataset):
            pass
        else:
            pass  # good luck :crossed-fingers:

        out.append(entry)

    return out


def plot(
    plot_type: str,
    /,
    *,
    data: list,
    style: dict = None,
    config: dict = None,
    size=(1280, 720),
    image_size=None,
    save_path=None,
    save_replace_existing=False,
    scale=1,
    watermark=False,
    watermark_position=DEFAULT,
    watermark_offset=DEFAULT,
    watermark_scale=DEFAULT,
):
    # whitelisted plots that share same "data" data structures, and can be normalised
    if plot_type in {"line", "scatter", "histogram", "joint_probability", "seasonality", "box", "rose"}:
        data = _normalise_data(data)

    if style is None:
        style = {}

    if config is None:
        config = {}

    width, height = size

    # NOTE: GUI plotter can't natively handle min and max axis
    # values that are dates. We can hack this in by converting dates
    # to just the epoch timestamp
    # (without the "_type": "datetime" wrapper)
    style = json.loads(json.dumps(style, cls=serializer.Encoder))
    for bound in ["max_x", "max_y", "min_x", "min_y"]:
        if (
            isinstance(style.get(bound), dict)
            and style[bound].get("_type") == "datetime"
        ):
            style[bound] = style[bound]["epoch"]

    with tempfile.TemporaryDirectory(prefix="te-plotting") as dir_:
        style_fp = os.path.join(dir_, "style.json")
        config_fp = os.path.join(dir_, "config.json")
        data_fp = os.path.join(dir_, "data.json")

        with open(style_fp, "w+") as fp:
            json.dump(style, fp, cls=serializer.Encoder)

        with open(config_fp, "w+") as fp:
            json.dump(config, fp, cls=serializer.Encoder)

        with open(data_fp, "w+") as fp:
            json.dump(data, fp, cls=serializer.Encoder)

        args = []

        if save_replace_existing:
            args.append("--override")

        kwargs = {
            "--plot_type": plot_type,
            "--style_file": style_fp,
            "--config_file": config_fp,
            "--data_file": data_fp,
            "--window-size": f"{width}x{height}",
            "--scale": scale,
        }

        if image_size:
            kwargs["--image-size"] = image_size

        if save_path:
            kwargs["--save"] = save_path

        args = [str(i) for i in itertools.chain(args, *kwargs.items())]
        try:
            _ = subprocess.check_output([cfg.plotter, *args])
        except subprocess.CalledProcessError as e:
            raise RuntimeError(e.stdout.decode("utf8"))

        if watermark and save_path:
            kwargs = dict(
                image_path=save_path,
                save_path=save_path,
                watermark_path=WATERMARK_PATH,
            )
            if watermark_position is not DEFAULT:
                kwargs["position"] = watermark_position
            if watermark_offset is not DEFAULT:
                kwargs["offset"] = watermark_offset
            if watermark_scale is not DEFAULT:
                kwargs["scale"] = watermark_scale

            _apply_watermark(**kwargs)



import os

from wave_venture import utils


url = os.environ.get("URL", "https://daemon.wave-venture.com/api")
plotter = os.environ.get("PLOTTER", utils.find_gui_plotter())
auth_token = os.environ.get("AUTH_TOKEN", utils.find_auth_token())

websocket_max_size = os.environ.get("WEBSOCKET_MAX_SIZE", None)
websocket_max_queue = os.environ.get("WEBSOCKET_MAX_QUEUE", None)
websocket_read_limit = os.environ.get("WEBSOCKET_READ_LIMIT", 2**32)
websocket_write_limit = os.environ.get("WEBSOCKET_WRITE_LIMIT", 2**32)
websocket_close_timeout = os.environ.get("WEBSOCKET_CLOSE_TIMEOUT", 10)
websocket_ping_interval = os.environ.get("WEBSOCKET_PING_INTERVAL", 20)
websocket_ping_timeout = os.environ.get("WEBSOCKET_PING_TIMEOUT", 20)



import os
import sys
import pytz
import json
import asyncio
import datetime
import functools


ALL_METATYPES = {
    "body",
}


def is_document(obj):
    return isinstance(obj, dict) and obj.get("_type") == "doc"


def assert_call_signature(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if len(args) == 0:
            raise ValueError(
                "insert must be called with a metatype name 'doc', 'body', etc as it's first positional argument."
            )

        _metatype = args[0]

        if _metatype == "doc" and len(args) == 1:
            return f(*args, **kwargs)
        elif _metatype in ALL_METATYPES and len(args) == 2 and is_document(args[1]):
            return f(*args, **kwargs)
        elif _metatype == "doc" and len(args) > 1:
            raise ValueError(
                'when inserting a doc, all arguments following "doc" must be keyword arguments. e.g. di.insert("doc", name="my_new_doc")'
            )
        elif _metatype not in ALL_METATYPES:
            raise ValueError(f"unknown prototype '{_metatype}'")
        elif len(args) > 2:
            raise ValueError(
                'when inserting a prototype, all arguments following the prototype name and the document must be keyword arguments. e.g. di.insert("body", doc, name="my_new_doc")'
            )
        else:
            raise ValueError("incorrect call")

    return wrapper


WIN_GUI_LOCAL_PATHS = (
    r"~\AppData\Local\Wave Venture\Wave Venture TE\main\local.json",
)
MAC_GUI_LOCAL_PATHS = (
    r"~/Library/Preferences/Wave Venture/Wave Venture TE/main/local.json",
)
LINUX_GUI_LOCAL_PATHS = (
    r"/etc/xdg/Wave Venture/Wave Venture TE/main/local.json",
    r"~/.config/Wave Venture/Wave Venture TE/main/local.json",
)


def find_auth_token():
    path_candidates = []
    if sys.platform.startswith("win"):
        path_candidates = WIN_GUI_LOCAL_PATHS
    elif sys.platform == "darwin":
        path_candidates = MAC_GUI_LOCAL_PATHS
    else:
        path_candidates = LINUX_GUI_LOCAL_PATHS

    for path in path_candidates:
        path = os.path.expanduser(path)
        try:
            with open(path, "r") as fp:
                local = json.load(fp)

            return local["server"]["token"]
        except (FileNotFoundError, json.JSONDecodeError, KeyError):
            pass


WIN_GUI_REGISTRY_PATH = (
    r"SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall\wv_gui_is1"
)
MAC_GUI_PLOTTER_PATH = "/Applications/Wave Venture TE.app/Contents/MacOS/wv_plotter"
LINUX_GUI_PLOTTER_PATH = "/usr/bin/wv_plotter"


def find_gui_plotter():
    if sys.platform.startswith("win"):
        import winreg

        possible_keys = [
            # HKEY_CURRENT_USER for non admin install
            # (like corporate computer without admin right)
            winreg.HKEY_CURRENT_USER,
            # - HKEY_LOCAL_MACHINE for admin installation
            # (most of home computer where the user is the admin)
            winreg.HKEY_LOCAL_MACHINE,
        ]

        for key in possible_keys:
            try:
                with winreg.OpenKey(key, WIN_GUI_REGISTRY_PATH) as key:
                    path, _ = winreg.QueryValueEx(key, "InstallLocation")
            except FileNotFoundError:
                continue

            path = os.path.join(path, "wv_plotter.exe")
            if os.path.exists(path):
                return path
    elif sys.platform == "darwin" and os.path.exists(MAC_GUI_PLOTTER_PATH):
        return MAC_GUI_PLOTTER_PATH
    elif os.path.exists(LINUX_GUI_PLOTTER_PATH):
        return LINUX_GUI_PLOTTER_PATH


def serial(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))
    return wrapper


EPOCH = datetime.datetime(1970, 1, 1)
UTC_EPOCH = EPOCH.replace(tzinfo=datetime.timezone.utc)


def to_epoch(dt):
    return (dt - EPOCH).total_seconds()


def from_epoch(seconds):
    return datetime.datetime.utcfromtimestamp(0) + datetime.timedelta(seconds=seconds)


def hours_between(d1, d2):
    return (d2 - d1).total_seconds() / 60 / 60


def is_date(d):
    if isinstance(d, datetime.date):
        return True
    elif isinstance(d, datetime.datetime):
        date_resolution = datetime.datetime(year=d.year, month=d.month, day=d.day)
        return d == date_resolution
    else:
        raise ValueError(f"Expecting datetime.datetime or datetime.date, got {type(d)}")


def is_naive(d):
    return not (d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None)


def is_utc(d):
    return d.tzinfo == datetime.timezone.utc or d.tzinfo == pytz.utc



import json
import datetime

import numpy as np
import pandas as pd
import xarray as xr

from wave_venture import utils


def _decode_array_values(data):
    _type = data["_type"]

    if _type in {"number", "fraction", "degree", "radian", "boolean"}:
        return data["values"]
    elif _type == "complex":
        return np.array(data["real"]) + np.array(data["imag"]) * 1j
    elif _type == "string":
        return data["values"]
    elif _type == "datetime":
        return pd.to_datetime(data["epochs"], unit="s")
    else:
        raise NotImplementedError(data)


def decoder(obj):
    _type = obj.get("_type", None)  # safely get _type if it exists

    if _type == "datetime" and "epoch" in obj:
        return utils.from_epoch(obj["epoch"])
    elif _type == "array":
        array = xr.Dataset(
            data_vars={
                name: (obj["dimensions"], _decode_array_values(data))
                for name, data in obj["fields"].items()
            },
            coords={
                name: _decode_array_values(data)
                for name, data in obj["indices"].items()
            },
        )
        for name in obj["indices"].keys():
            array[name].attrs["interpolation"] = obj["indices"][name].get("interpolation")
            array[name].attrs["_type"] = obj["indices"][name]["_type"]
        for name in obj["fields"].keys():
            array[name].attrs["_type"] = obj["fields"][name]["_type"]
        return array
    else:
        return obj


_TYPE_CHECKS = {
    "number": lambda x: np.issubdtype(x.dtype, np.number),
    "fraction": lambda x: np.issubdtype(x.dtype, np.number),
    "degree": lambda x: np.issubdtype(x.dtype, np.number),
    "radian": lambda x: np.issubdtype(x.dtype, np.number),
    "datetime": lambda x: x.dtype == np.dtype("<M8[ns]"),
    "string": lambda x: pd.api.types.is_string_dtype(x),
    "boolean": lambda x: np.issubdtype(x, np.dtype("bool")),
    "complex": lambda x: np.issubdtype(x, np.dtype("complex")),
}

_TYPE_ENCODER = {
    "number": lambda x: {"values": x.to_numpy().tolist()},
    "fraction": lambda x: {"values": x.to_numpy().tolist()},
    "degree": lambda x: {"values": x.to_numpy().tolist()},
    "radian": lambda x: {"values": x.to_numpy().tolist()},
    "datetime": lambda x: {"epochs": (x.to_numpy().view(np.int64) / 10 ** 9).tolist()},
    "string": lambda x: {"values": list(map(str, x.to_numpy().tolist()))},
    "boolean": lambda x: {"values": x.to_numpy().tolist()},
    "complex": lambda x: {"real": x.real.to_numpy().tolist(), "imag": x.imag.to_numpy().tolist()},
}


def _encode_array_values(xarray, *, name):
    # get _type
    if (
        name in xarray.data_vars
        and "_type" in xarray.data_vars[name].attrs
    ):
        _type = xarray.data_vars[name].attrs["_type"]
    elif (
        name in xarray.coords
        and "_type" in xarray.coords[name].attrs
    ):
        _type = xarray.coords[name].attrs["_type"]
    else:
        _type = None

    if _type is None:
        # NOTE: order important here, as first match wins the inference battle
        inferable_type = ["datetime", "complex", "boolean", "number", "string"]
        for candidate_type in inferable_type:
            if _TYPE_CHECKS[candidate_type](xarray[name]):
                _type = candidate_type
                break

    if _type is None:
        raise ValueError(
            f"Unable to find associated _type for '{name}' "
            f"or able to infer a _type from its dtype '{xarray[name].dtype}'. "
        )

    if not _TYPE_CHECKS[_type](xarray[name]):
        raise ValueError(f"Values for '{name}' dont look like '{_type}'")

    result = _TYPE_ENCODER[_type](xarray[name])
    result = {**result, "_type": _type}

    # preserve interpolation attrs if exist
    if name in set(xarray.coords):
        if interpolation := xarray[name].attrs.get("interpolation"):
            result["interpolation"] = interpolation

    return result


class Encoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "shape") and obj.shape == ():
            # unwraps zero-shape scalars like np.array(1) into 1
            return obj.item()
        elif isinstance(obj, datetime.datetime) and utils.is_naive(obj):
            return {
                "_type": "datetime",
                "epoch": utils.to_epoch(obj),
            }
        elif isinstance(obj, datetime.datetime) and utils.is_utc(obj):
            return {
                "_type": "datetime",
                "epoch": utils.to_epoch(obj),
            }
        elif isinstance(obj, datetime.datetime):
            raise TypeError("Encountered unexpected timezone aware datetime.")
        elif isinstance(obj, np.datetime64):
            return self.default(pd.to_datetime(obj).to_pydatetime())
        elif isinstance(obj, pd.Timestamp):
            return self.default(obj.to_pydatetime())
        elif isinstance(obj, datetime.date):
            return self.default(datetime.datetime(obj.year, obj.month, obj.day))
        elif isinstance(obj, complex):
            return {"_type": "complex", "real": obj.real, "imag": obj.imag}
        # just those arrays that can be a scalar,
        # i.e. np.array(1), pd.DataArray(1)
        elif isinstance(obj, (np.ndarray, xr.DataArray)) and obj.shape == ():
            if np.issubdtype(obj.dtype, np.datetime64):
                # calling .item on a datetime scalar will return its
                # epoch float, we dont want that, we want a datetime
                # we can encode appropriately with _type: datetime.
                if hasattr(obj, "to_numpy"):
                    obj = obj.to_numpy()
                scalar = pd.to_datetime(obj).to_pydatetime()
            else:
                scalar = obj.item()
            return self.default(scalar)
        elif isinstance(obj, np.ndarray):
            return np.vectorize(self.default)(obj).tolist()
        elif isinstance(obj, (xr.Dataset, xr.DataArray, pd.Series, pd.DataFrame)):
            if isinstance(obj, xr.DataArray):
                obj = obj.to_dataset(name="unnamed")
            elif isinstance(obj, pd.Series):
                obj = obj.to_xarray().to_dataset(name="unnamed")
            elif isinstance(obj, pd.DataFrame):
                obj = obj.to_xarray()

            return {
                "_type": "array",
                "fields": {
                    column: _encode_array_values(obj, name=column)
                    for column in obj.data_vars
                },
                "dimensions": list(obj.coords),
                "indices": {
                    coord: _encode_array_values(obj, name=coord)
                    for coord in list(obj.coords)
                },
            }
        else:
            return json.JSONEncoder.default(self, obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os

from setuptools import find_packages, setup

# Package meta-data.
NAME = "wave-venture"
DESCRIPTION = "Wave Venture's Python interface to TE Software API."
URL = "https://github.com/waveventure/wv_py"
EMAIL = "admin@wave-venture.com"
AUTHOR = "Wave Venture"
REQUIRES_PYTHON = ">=3.8.0"
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    "certifi",
    "loguru",
    "numpy",
    "pandas",
    "pillow",
    "pytz",
    "requests",
    "scipy",
    "tqdm",
    "websockets",
    "xarray",
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!
ROOT = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(ROOT, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(ROOT, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/plain",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["test", "*.test", "*.test.*", "test.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # $ setup.py publish support.
    cmdclass={
        # 'upload': UploadCommand,
        # "build_ext": Builder,
    },
)



import ssl
import json
import uuid
import base64
import typing
import hashlib
import urllib.parse

import tqdm
import loguru
import certifi
import requests
import websockets

from wave_venture import utils
from wave_venture import config
from wave_venture import serializer


class ServerError(Exception):
    pass


def make_headers() -> dict:
    headers = {}

    if config.auth_token:
        headers["Authorization"] = f"Bearer {config.auth_token}"

    return headers


def _send(action, kwargs):
    response = requests.post(
        url=(config.url + "/" + action),
        json=kwargs,
        headers=make_headers(),
        verify=certifi.where(),
    )
    response.raise_for_status()
    body = response.json(object_hook=serializer.decoder)
    if body["error"]:
        if trace := body["error"].get("stack_trace"):
            loguru.logger.error(trace)
        raise ServerError(body["error"]["dev_message"])
    return body["result"]


async def _send_progress(action, kwargs, *, progress_title):
    url = urllib.parse.urlparse(config.url)
    secure = url.scheme == "https"
    scheme = "wss" if secure else "ws"
    url = url._replace(scheme=scheme)
    url = urllib.parse.urlunparse(url)

    req_uid = uuid.uuid4().hex
    ssl_ctx = ssl.create_default_context(cafile=certifi.where())
    ssl_ctx = ssl_ctx if secure else None
    headers = make_headers()

    connection = websockets.connect(
        uri=url,
        extra_headers=headers,
        ssl=ssl_ctx,
        max_size=config.websocket_max_size,
        max_queue=config.websocket_max_queue,
        read_limit=config.websocket_read_limit,
        write_limit=config.websocket_write_limit,
        close_timeout=config.websocket_close_timeout,
        ping_interval=config.websocket_ping_interval,
        ping_timeout=config.websocket_ping_timeout,
    )

    async with connection as ws:
        payload = json.dumps(
            {
                "uid": req_uid,
                "action": action,
                "kwargs": kwargs,
            },
        )
        await ws.send(payload)

        progress_bar = tqdm.tqdm(
            total=100,
            desc=progress_title,
            bar_format="{l_bar}{bar}",
        )
        while True:
            msg = await ws.recv()
            msg = json.loads(msg, object_hook=serializer.decoder)

            if msg["error"]:
                if trace := msg["error"].get("stack_trace"):
                    loguru.logger.error(trace)
                raise ServerError(msg["error"]["dev_message"])

            assert msg["uid"] == req_uid

            if (
                msg["result"] is not None
                and isinstance(msg["result"], dict)
                and msg["result"].get("_type") == "progress"
            ):
                # progress_bar.set_description(msg["result"]["message"])
                update = (msg["result"]["progress"] * 100) - progress_bar.n
                progress_bar.update(int(update))
                yield msg["result"]
            else:
                progress_bar.set_description("complete")
                progress_bar.update(100 - progress_bar.n)
                break

        progress_bar.clear()
        progress_bar.close()

    yield msg["result"]


# @utils.assert_call_signature
def new(
    *,
    name: str,
    prescript: str = None,
    tags_pre: typing.List[str] = None,
) -> dict:
    result = _send(
        action="document.insert",
        kwargs={
            "name": name,
            "prescript": prescript,
            "tags_pre": tags_pre,
        },
    )
    return result["instance"]


def load(*, uid: str) -> dict:
    result = _send(
        action="document.select",
        kwargs={"uid": uid},
    )
    return result["instance"]


@utils.serial
async def import_(export_path):
    with open(export_path, "rb") as fp:
        data = fp.read()

    runner = _send_progress(
        action="document.import",
        kwargs={
            "archive": {
                "name": "document.zip",
                "sha256": hashlib.sha256(data).hexdigest(),
                "bytes": base64.urlsafe_b64encode(data).decode("utf8"),
            },
        },
        progress_title="importing",
    )

    async for result in runner:
        pass

    return result[0]


def clone(
    *,
    uid: str,
    name: str,
    prescript: str = None,
    tags_pre: typing.List[str] = None,
    trim_errors: bool = False,
) -> dict:
    result = _send(
        action="document.clone",
        kwargs={
            "previous_uid": uid,
            "new_name": name,
            "new_prescript": prescript,
            "new_tags_pre": tags_pre,
            "trim_errors": trim_errors,
        },
    )
    return result["instance"]


@utils.serial
async def run(doc: dict) -> dict:
    action = "document.update"
    kwargs = {"uid": doc["uid"], "finalised": True}
    runner = _send_progress(action, kwargs, progress_title="running")

    async for result in runner:
        pass

    return result["instance"]


def equation(value: str):
    return {"_type": "equation", "value": str(value)}


# @utils.assert_call_signature
def add(
    metatype: str,
    doc: dict,
    /,
    *,
    name: str,
    **kwargs,
) -> dict:
    result = _send(
        action="prototype.insert",
        kwargs={
            "doc_uid": doc["uid"],
            "instance": {
                "name": name,
                "_metatype": metatype,
                **kwargs,
            },
        },
    )
    return result["instance"]


# @utils.assert_call_signature
def update(
    metatype: str,
    doc: dict,
    /,
    **kwargs,
) -> dict:
    result = _send(
        action="prototype.update",
        kwargs={
            "doc_uid": doc["uid"],
            "instance": {
                "_metatype": metatype,
                **kwargs,
            },
        },
    )
    return result["instance"]


@utils.serial
async def resolve(
    doc_or_perms,
    results_paths,
) -> typing.List[dict]:
    if (
        isinstance(doc_or_perms, dict)
        and doc_or_perms.get("_type") == "doc"
    ):
        result = _send(
            action="document.select",
            kwargs={
                "uid": doc_or_perms["uid"],
                "include": ["uid", "finalised"],
            },
        )
        doc = result["instance"]

        if doc["finalised"]:
            permutations = _send(
                action="permutation.select",
                kwargs={
                    "doc_uid": doc["uid"],
                },
            )
            uids = [p["uid"] for p in permutations if p["success"] is True]
        else:
            uids = [doc["uid"]]

    elif (
        isinstance(doc_or_perms, list)
        and all(isinstance(p, dict) and p.get("_type") == "permutation" for p in doc_or_perms)
    ):
        uids = [p["uid"] for p in doc_or_perms]
    elif (
        isinstance(doc_or_perms, list)
        and all(isinstance(p, str) and (p.startswith("doc_") or p.startswith("perm_")) for p in doc_or_perms)
    ):
        uids = doc_or_perms
    else:
        raise ValueError("expected document or list of permutations")

    if isinstance(results_paths, str):
        results_paths = [
            segment.strip()
            for line in results_paths.split("\n")
            for segment in line.split(" ")
            if segment
        ]

    results_paths = ["uid", *results_paths]

    action = "resolve"
    kwargs = {"uids": uids, "paths": results_paths}
    runner = _send_progress(action, kwargs, progress_title="resolving")

    results = {}

    async for result in runner:
        if result and result.get("_type") == "progress":
            uid = result["chunk"]["uid"]
            path = result["chunk"]["path"]

            if error := result["chunk"].get("error"):
                if trace := error.get("stack_trace"):
                    loguru.logger.error(trace)

                raise ServerError(
                    f"Error while resolving '{path}': {error['user_message']}",
                )
            else:
                value = result["chunk"]["value"]
                results.setdefault(uid, {})[path] = value

    # sort in order of uids
    return [results[uid] for uid in uids]



from wave_venture.interface import new, load, clone, import_, add, update, run, resolve
from wave_venture.plotting import plot

__all__ = ["new", "load", "clone", "import_", "add", "update", "run", "resolve", "plot"]
